reset()

import numpy
from numpy import linalg
numpy.random.seed(60856)

import scipy
from scipy.stats.mstats import mquantiles

import matplotlib
matplotlib.use('SVG')
matplotlib.rc('font',family='M+ 2p',weight='light')
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse

from sage.misc.decorators import infix_operator
@infix_operator('multiply')
def dot(a,b) : return vector([a[i]*b[i] for i in range(0,min(len(a),len(b)))])

if (numpy.random.randint(10000) != 7990) : print("Warning: Your random number generator did not generate the same number as mine, results may vary")

var('theta x0 x1 x2')
print('State Vector : [cos(theta),sin(theta),Dtheta]=[x0,x1,x2]')
print('Which makes the continuous time update equation')
print('x0 = cos(theta+Dtheta*t) = cos(theta)*cos(Dtheta*t) - sin(theta)*sin(Dtheta*t) = x0*cos(x2*t) - x1*sin(x2*t)')
print('x1 = sin(theta+Dtheta*t) = cos(theta)*sin(Dtheta*t) + sin(theta)*cos(Dtheta*t) = x0*sin(x2*t) + x1*cos(x2*t)')
print('x2 = x2')
print('Assume x2*t~0')
print('x0 = x0-x1*x2*t')
print('x1 = x1+x0*x2*t')
print('x2 = x2')

dt = 0.2
def f(x,w) : return vector([x[0]-x[1]*x[2]*dt,x[1]+x[0]*x[2]*dt,x[2]+w[0]])
var('x0 x1 x2 w0')
fe = f([x0,x1,x2],[w0])
F = matrix([
    [diff(fe[0],x0),diff(fe[0],x1),diff(fe[0],x2)],
    [diff(fe[1],x0),diff(fe[1],x1),diff(fe[1],x2)],
    [diff(fe[2],x0),diff(fe[2],x1),diff(fe[2],x2)]
]).subs(w0=0)
L = matrix([
    [diff(fe[0],w0)],
    [diff(fe[1],w0)],
    [diff(fe[2],w0)]
])
Q_root = matrix([[0.4]])
Q = Q_root*Q_root.transpose()

H = matrix([0,1,0])
J = matrix([[1]])
R_root = matrix([[.25]])
R = R_root*R_root.transpose()
def obs(x,v) : return H*x+J*v

npoints = 200
num_steps = 30

rand_itheta = (2*pi*(numpy.random.uniform(0,.0)+0)).n()

s0h = vector([cos(rand_itheta),sin(rand_itheta),.0])
small_tform = matrix([
    [1,0,0,-s0h[1]],
    [0,1,0,-s0h[0]],
    [0,0,1,0]
])
P0 = small_tform*diagonal_matrix([0,0,.3,.3])*small_tform.transpose()
tmp = linalg.svd(P0)
P0_root = matrix(tmp[0])*diagonal_matrix(vector(tmp[1]).apply_map(sqrt))*matrix(tmp[2])

#Define the initial state mean
print("The initial state")
print("    Mean")
print(s0h.n(digits=3))
print("    Cov")
print(P0.n(digits=3))
print("F")
print(F)

def create_realization() : 
    s0 = s0h + P0_root*vector(numpy.random.normal(0,1,3))
    w_noise = [Q_root*vector(numpy.random.normal(0,1,1)) for k in range(0,num_steps)]
    v_noise = [R_root*vector(numpy.random.normal(0,1,1)) for k in range(0,num_steps)]
    truth_state = [s0 for k in range(0,num_steps+1)]
    time = [0 for k in range(0,num_steps+1)]
    for k in range(1,num_steps+1) :
        truth_state[k] = f(truth_state[k-1],w_noise[k-1])
        time[k] = time[k-1]+dt
    z = [obs(truth_state[k],v_noise[k-1]) for k in range(1,num_steps+1)]
    return (truth_state,w_noise,v_noise,time,z)

realizations = [create_realization() for _ in range(0,npoints)]

load("Generate_Sigma_Points.sage")
#import Generate_Sigma_points

Simplex_set = Generate_Simplex(4).n()
Order3_nopair = Generate_Order3_nopair(4).n()
Order5_nopair = Generate_Order5_nopair(4).n()
Order7_nopair = Generate_Order7_nopair(4).n()
Order9_nopair = Generate_Order9_nopair(4).n()
Order4_pair = Generate_Order4_pair(4).n()

spoint_sets = [
    (Simplex_set,"UKF_simp","UKF (Simp)","#ff9900",'^--'),
    (Order3_nopair,"UKF_O3","UKF (O3)","#00a250",'v--'),
    (Order5_nopair,"UKF_O5","UKF (O5)","#336633",'v-'),
    (Order7_nopair,"UKF_O7","UKF (O7)","#00a9e4",'^--'),
    (Order9_nopair,"UKF_O9","UKF (O9)","#336699",'v-'),
    (Order4_pair,"UKF_O4p","UKF (O5f)","#993300",'v-')
]

def run_linearized(ss0h,zz,lin_points) :
    ss = [ss0h for k in range(0,num_steps+1)]
    pp = [P0 for k in range(0,num_steps+1)]
    for k in range(1,num_steps+1) :
        s = ss[k-1]
        p = pp[k-1]
        lp = lin_points[k-1]
        Fe = F.subs({x0:lp[0],x1:lp[1],x2:lp[2]})
        sn = f(lp,[0])+Fe*(s-lp)
        pn = Fe*p*Fe.transpose()+L*Q*L.transpose()
        K = pn*H.transpose()*(H*pn*H.transpose()+J*R*J.transpose()).inverse()
        tr = zz[k-1]-obs(sn,vector([0])) #We can do this because obs is linear
        su = sn + K*tr
        tt = identity_matrix(3)-K*H
        pu = tt*pn*tt.transpose()+K*J*R*J.transpose()*K.transpose()
        ss[k] = su
        pp[k] = pu
    return(ss,pp)

def run_ekf(ss0h,zz) :
    ss = [ss0h for k in range(0,num_steps+1)]
    pp = [P0 for k in range(0,num_steps+1)]
    for k in range(1,num_steps+1) :
        s = ss[k-1]
        p = pp[k-1]
        lp = s
        Fe = F.subs({x0:lp[0],x1:lp[1],x2:lp[2]})
        sn = f(lp,[0])+Fe*(s-lp)
        pn = Fe*p*Fe.transpose()+L*Q*L.transpose()
        K = pn*H.transpose()*(H*pn*H.transpose()+J*R*J.transpose()).inverse()
        tr = zz[k-1]-obs(sn,vector([0])) #We can do this because obs is linear
        su = sn + K*tr
        tt = identity_matrix(3)-K*H
        pu = tt*pn*tt.transpose()+K*J*R*J.transpose()*K.transpose()
        ss[k] = su
        pp[k] = pu
    return(ss,pp)

def eval_spoint(ef,s,p_r,v,r_r,spoints) :
    weights = spoints[0]
    spoints_t = spoints.submatrix(1,0).transpose()
    ty = [ef(s,v) for _ in range(0,len(weights))]
    ts = [s for _ in range(0,len(weights))]
    for j in range(0,len(weights)) :
        tsp = spoints_t[j]
        ts[j] = s + p_r*vector([tsp[0],tsp[1],tsp[2]])
        tv = v + r_r*vector([tsp[3]])
        ty[j] = ef(ts[j],tv).n()
    ym = sum([weights[j]*ty[j] for j in range(0,len(weights))])
    pzz = sum([weights[j]*(ty[j]-ym).outer_product(ty[j]-ym) for j in range(0,len(weights))])
    pxz = sum([weights[j]*(ts[j]-s).outer_product(ty[j]-ym) for j in range(1,len(weights))])
    return (ym,pzz,pxz)

def run_spoint(ss0h,zz,spoints) :
    ss = [ss0h for k in range(0,num_steps+1)]
    pp = [P0 for k in range(0,num_steps+1)]
    for k in range(1,num_steps+1) :
        s = ss[k-1]
        p = pp[k-1]
        tmp = linalg.svd(p)
        p_root = matrix(tmp[0])*diagonal_matrix(vector(tmp[1]).apply_map(sqrt))*matrix(tmp[2])
        (sn,pn,pnp) = eval_spoint(f,s,p_root,vector([0]),Q_root,spoints)
        K = pn*H.transpose()*(H*pn*H.transpose()+J*R*J.transpose()).inverse()
        tr = zz[k-1]-obs(sn,vector([0])) #We can do this because obs is linear
        su = sn + K*tr
        tt = identity_matrix(3)-K*H
        pu = tt*pn*tt.transpose()+K*J*R*J.transpose()*K.transpose()
        ss[k] = su
        pp[k] = pu
    return(ss,pp)

def generate_xt(time,truth,s_fil_p,s_fil,name,col,filename,full_label) :
    fig = plt.figure()
    fig.set_figwidth(3)
    fig.set_figheight(2.5)
    axs = fig.add_axes( (0.6/7,0.4/4,1-.8/7,1-.6/4))

    axs.set_xlabel("Time")
    axs.set_ylabel("Theta")

    tru, = axs.plot(time,map(lambda u : u[0],truth),'k-',linewidth=3.5)
    per, = axs.plot(time,map(lambda u : u[0],s_fil_p),'^-',color='#8F00FF',linewidth=.7,markeredgecolor='k',markeredgewidth=0.7,markersize=6)
    fil, = axs.plot(time,map(lambda u : u[0],s_fil),'v--',color=col,linewidth=1.5,markeredgecolor='k',markersize=5,markeredgewidth=0.5)

    if full_label : 
        axs.legend( (tru,per,fil), ("Truth","Perfect (KF)",name), framealpha=0.7,numpoints=1,prop={'family':'M+ 2p','weight':'ultralight','size':8})
    else : 
        axs.legend( [fil], [name], framealpha=0.7,numpoints=1,prop={'family':'M+ 2p','weight':'ultralight','size':8})

    yticks = [0,1,-1]
    yticks.sort()
    axs.set_yticks(yticks)
    axs.set_yticklabels(map(lambda u : u.n(digits=2),yticks), fontsize=8, family='sans-serif', weight='light')
    axs.set_xticks([])
    #axs.set_xticks(time,minor=True)
    axs.grid(True)

    fig.savefig('Plots/Angle_Time_'+filename+'.svg')
    plt.close(fig)

clist = [
    '#cc0000',
    '#ef5f21',
    '#00a250',
    '#336633',
    '#00a9e4',
    '#336699']

#Run the first set
this_run = realizations[0]
time = this_run[3]
(s_per,p_per) = run_linearized(s0h,this_run[4],this_run[0])
(s_ekf,p_ekf) = run_ekf(s0h,this_run[4])

generate_xt(this_run[3],this_run[0],s_per,s_ekf,"EKF",clist[0],"EKF",True)
for k in range(0,len(spoint_sets)) :
    (s_sp,p_sp) = run_spoint(s0h,this_run[4],spoint_sets[k][0])
#    generate_xt(this_run[3],this_run[0],s_per,s_sp,spoint_sets[k][2],clist[k+1],spoint_sets[k][1],False)

err_per = [ [] for k in range(0,npoints)]
err_ekf = [ [] for k in range(0,npoints)]
err_sp = [ [ [] for k in range(0,npoints)] for j in range(0,len(spoint_sets))]

def find_err(truth,fil) :
    d_err = [vector([truth[k][0] - fil[k][0],truth[k][1]-fil[k][1]]) for k in range(0,len(fil))]
    return map(lambda u : sqrt(u*u), d_err)

for ind in range(0,npoints) :
    this_run = realizations[ind]
    (s_per,p_per) = run_linearized(s0h,this_run[4],this_run[0])
    err_per[ind] = find_err(this_run[0],s_per)
    (s_ekf,p_ekf) = run_ekf(s0h,this_run[4])
    err_ekf[ind] = find_err(this_run[0],s_ekf)
    for k in range(0,len(spoint_sets)) :
        (s_sp,p_sp) = run_spoint(s0h,this_run[4],spoint_sets[k][0])
        err_sp[k][ind] = find_err(this_run[0],s_sp)


def add_plot(q,td,col,axs,sym) :
    tt = map(lambda u : u+td,time)
    qp = axs.plot(tt,list(q),sym,color=col,linewidth=1.7,markeredgecolor='k',markeredgewidth=0.7,markersize=6)
    return(qp)

def draw_quartile(qt,name,ybnd,ytk) :
    quant_per = mquantiles(err_per,prob=[qt],axis=0)[0]
    quant_ekf = mquantiles(err_ekf,prob=[qt],axis=0)[0]
    quant_sp = [ mquantiles(err_sp[k],prob=[qt],axis=0)[0] for k in range(0,len(spoint_sets))]
    
    fig = plt.figure()
    fig.set_figwidth(7)
    fig.set_figheight(5)
    axs = fig.add_axes( (0.6/7,0.4/4,1-.8/7,1-.6/4) )

    axs.set_xlabel("Time")
    axs.set_ylabel("Error")

    per, = add_plot(quant_per,0,'#8F00FF',axs,'v-')
    ekf, = add_plot(quant_ekf,0.0,'#cc0000',axs,'v-')
    
    leg_entries = [ per for k in range(0,len(spoint_sets)+2)]
    leg_names = ["" for k in range(0,len(spoint_sets)+2)]
    leg_entries[0] = per
    leg_names[0] = "Perfect (KF)"
    leg_entries[1] = ekf
    leg_names[1] = "EKF"

    for k in range(0,len(spoint_sets)) :
        sp, = add_plot(quant_sp[k],0.0,spoint_sets[k][3],axs,spoint_sets[k][4])
        leg_entries[k+2] = sp
        leg_names[k+2] = spoint_sets[k][2]
    axs.legend( tuple(leg_entries) , tuple(leg_names),
                framealpha=0.7,numpoints=1,prop={'family':'M+ 2p','weight':'ultralight','size':10})
    axs.set_xticks( map(lambda u : u*dt,[0,10,20,30]) )
    yticks = ytk # + [quant_ekf[num_steps],quant_per[num_steps]] + [quant_sp[k][num_steps] for k in range(0,len(quant_sp))]
    yticks.sort()
    axs.set_yticks( map(lambda u : n(u,digits=3),yticks) )
    axs.set_ylim( ybnd )
    axs.set_xlim( (0,6) )
    axs.grid(True)
    
    fig.savefig('Plots/Angle_Error_'+name+'.svg')
    plt.close(fig)

draw_quartile(.5,'50th',(0,2),[0,0.5,1,1.5,2])
draw_quartile(.75,'75th',(0,4),[0,1,2,3,4])
draw_quartile(.90,'90th',(0,8),[0,2,4,6,8])
#draw_quartile(1,'max')
